from __future__ import print_function, unicode_literals

import mimetypes
import os


class RequirementsDetector(object):
    """ Takes raw requirements argument, and detects / discovers all the requirements files. """

    filenames = []

    def __init__(self, requirements_arg):
        self.filenames = []

        if not requirements_arg:
            self.autodetect_files()
        else:
            self.detect_files(requirements_arg)

    def get_filenames(self):
        """ Returns a list of all filenames detected as proper requirements files. """
        return self.filenames

    def detect_files(self, requirements_arg):
        for argument in requirements_arg:
            if self._is_valid_requirements_file(argument):
                self.filenames.append(argument)
            else:  # pragma: nocover
                print('Invalid requirements file: {}'.format(argument))

    def autodetect_files(self):
        """ Attempt to detect requirements files in the current working directory """
        if self._is_valid_requirements_file('requirements.txt'):
            self.filenames.append('requirements.txt')

        if self._is_valid_requirements_file('requirements.pip'):  # pragma: nocover
            self.filenames.append('requirements.pip')

        if os.path.isdir('requirements'):
            for filename in os.listdir('requirements'):
                file_path = os.path.join('requirements', filename)
                if self._is_valid_requirements_file(file_path):
                    self.filenames.append(file_path)

    @staticmethod
    def _is_valid_requirements_file(filename):
        extension_ok = filename.endswith('txt') or filename.endswith('pip')
        return extension_ok and os.path.isfile(filename) and mimetypes.guess_type(filename)[0] in ['text/plain', None]



from __future__ import print_function, unicode_literals

import subprocess
from subprocess import CalledProcessError

import re
from colorclass import Color


class PackagesUpgrader(object):

    selected_packages = None
    requirements_files = None
    upgraded_packages = None
    dry_run = False

    def __init__(self, selected_packages, requirements_files, options):
        self.selected_packages = selected_packages
        self.requirements_files = requirements_files
        self.upgraded_packages = []
        self.dry_run = options['--dry-run']
        self.skip_package_installation = options.get('--skip-package-installation', False)

    def do_upgrade(self):
        for package in self.selected_packages:
            self._update_package(package)

        return self.upgraded_packages

    def _update_package(self, package):
        """ Update (install) the package in current environment, and if success, also replace version in file """
        try:
            if not self.dry_run and not self.skip_package_installation:  # pragma: nocover
                subprocess.check_call(['pip', 'install', '{}=={}'.format(package['name'], package['latest_version'])])
            else:
                print('[Dry Run]: skipping package installation:', package['name'])
            # update only if installation success
            self._update_requirements_package(package)

        except CalledProcessError:  # pragma: nocover
            print(Color('{{autored}}Failed to install package "{}"{{/autored}}'.format(package['name'])))

    def _update_requirements_package(self, package):
        for filename in set(self.requirements_files):
            lines = []

            # read current lines
            with open(filename, 'r') as frh:
                for line in frh:
                    lines.append(line)

            try:
                # write updates lines
                with open(filename, 'w') as fwh:
                    for line in lines:
                        line = self._maybe_update_line_package(line, package)
                        fwh.write(line)
            except Exception as e:  # pragma: nocover
                # at exception, restore old file
                with open(filename, 'w') as fwh:
                    for line in lines:
                        fwh.write(line)
                raise e

    def _maybe_update_line_package(self, line, package):
        original_line = line
        pattern = r'\b{package}(?:\[\w*\])?=={old_version}\b'.format(
            package=re.escape(package['name']),
            old_version=re.escape(str(package['current_version'])))

        if re.search(pattern, line, flags=re.IGNORECASE):
            line = line.replace(
                '=={}'.format(package['current_version']),
                '=={}'.format(package['latest_version'])
            )

        if line != original_line:
            self.upgraded_packages.append(package)

            if self.dry_run:  # pragma: nocover
                print('[Dry Run]: skipping requirements replacement:',
                      original_line.replace('\n', ''), ' / ',
                      line.replace('\n', ''))
                return original_line
        return line



from __future__ import print_function, unicode_literals


class PackagesDetector(object):
    """ Takes list of requirements fies and returns the list of packages from all of them """

    packages = []

    def __init__(self, requirements_files):
        self.packages = []
        self.detect_packages(requirements_files)

    def get_packages(self):
        return self.packages

    def detect_packages(self, requirements_files):
        for filename in requirements_files:
            with open(filename) as fh:
                for line in fh:
                    self._process_req_line(line)

    def _process_req_line(self, line):

        if not line or not line.strip():
            return
        line = line.strip()

        if line.startswith('#'):
            return

        if line.startswith('-f') or line.startswith('--find-links') or \
                line.startswith('-i') or line.startswith('--index-url') or \
                line.startswith('--extra-index-url') or \
                line.startswith('--no-index') or \
                line.startswith('-r') or \
                line.startswith('-Z') or line.startswith('--always-unzip'):
            # private repositories
            return

        if '#' in line:  # inline comment in file
            line = line.split('#')[0].strip()

        self.packages.append(line)



from __future__ import print_function, unicode_literals

import os

import re
import requests
import sys

from colorclass import Color
from packaging import version
from packaging.utils import canonicalize_name
from pip.locations import site_config_files
from requests import HTTPError

try:
    from configparser import ConfigParser, NoOptionError
except ImportError:   # pragma: nocover
    from ConfigParser import ConfigParser, NoOptionError

try:
    from urllib.parse import urljoin
except ImportError:  # pragma: nocover
    from urlparse import urljoin


class PackagesStatusDetector(object):
    packages = []
    packages_status_map = {}
    PYPI_API_URL = None
    PYPI_API_TYPE = None
    pip_config_locations = [
        '~/.pip/pip.conf',
        '~/.pip/pip.ini',
        '~/.config/pip/pip.conf',
        '~/.config/pip/pip.ini',
    ]

    _prerelease = False

    def __init__(self, packages, use_default_index=False):
        self.packages = packages
        self.packages_status_map = {}
        self.PYPI_API_URL = 'https://pypi.python.org/pypi/{package}/json'
        self.PYPI_API_TYPE = 'pypi_json'

        if not use_default_index:
            self._update_index_url_from_configs()

        self._prerelease = False

    def _update_index_url_from_configs(self):
        """ Checks for alternative index-url in pip.conf """

        if 'VIRTUAL_ENV' in os.environ:
            self.pip_config_locations.append(os.path.join(os.environ['VIRTUAL_ENV'], 'pip.conf'))
            self.pip_config_locations.append(os.path.join(os.environ['VIRTUAL_ENV'], 'pip.ini'))

        if site_config_files:
            self.pip_config_locations.extend(site_config_files)

        index_url = None
        custom_config = None
        for pip_config_filename in self.pip_config_locations:
            if pip_config_filename.startswith('~'):
                pip_config_filename = os.path.expanduser(pip_config_filename)

            if os.path.isfile(pip_config_filename):
                config = ConfigParser()
                config.read([pip_config_filename])
                try:
                    index_url = config.get('global', 'index-url')
                    custom_config = pip_config_filename
                    break  # stop on first detected, because config locations have a priority
                except NoOptionError:  # pragma: nocover
                    pass

        if index_url:
            self.PYPI_API_URL = self._prepare_api_url(index_url)
            print(Color('Setting API url to {{autoyellow}}{}{{/autoyellow}} as found in {{autoyellow}}{}{{/autoyellow}}'
                        '. Use --default-index-url to use pypi default index'.format(self.PYPI_API_URL, custom_config)))

    def _prepare_api_url(self, index_url):  # pragma: nocover
        if not index_url.endswith('/'):
            index_url += '/'

        if index_url.endswith('/simple/'):
            self.PYPI_API_TYPE = 'simple_html'
            return urljoin(index_url, '{package}')

        if index_url.endswith('/+simple/'):
            self.PYPI_API_TYPE = 'simple_html'
            return urljoin(index_url, '{package}')

        if '/pypi/' in index_url:
            base_url = index_url.split('/pypi/')[0]
            return urljoin(base_url, '/pypi/{package}/json')

        return urljoin(index_url, '/pypi/{package}/json')

    def detect_available_upgrades(self, options):
        self._prerelease = options.get('--prerelease', False)
        explicit_packages_lower = None
        if options['-p'] and options['-p'] != ['all']:
            explicit_packages_lower = [pack_name.lower() for pack_name in options['-p']]

        for i, package in enumerate(self.packages):

            package_name, pinned_version = self._expand_package(package)
            if not package_name or not pinned_version:  # pragma: nocover
                # todo: treat <= or >= instead of ==
                continue

            if explicit_packages_lower and package_name.lower() not in explicit_packages_lower:
                # skip if explicit and not chosen
                continue

            current_version = version.parse(pinned_version)

            if pinned_version and isinstance(current_version, version.Version):  # version parsing is correct
                package_status, reason = self._fetch_index_package_info(package_name, current_version)
                if not package_status:  # pragma: nocover
                    print(reason)
                    continue

                print('{}/{}: {} ... '.format(i + 1, len(self.packages), package_name), end='')
                sys.stdout.flush()

                # compare versions
                if current_version < package_status['latest_version']:
                    print('upgrade available: {} ==> {} (uploaded on {})'.format(current_version,
                                                                                 package_status['latest_version'],
                                                                                 package_status['upload_time']))
                else:
                    print('up to date: {}'.format(current_version))
                sys.stdout.flush()

                self.packages_status_map[package_name] = package_status

        return self.packages_status_map

    def _fetch_index_package_info(self, package_name, current_version):
        """
        :type package_name: str
        :type current_version: version.Version
        """

        try:
            package_canonical_name = package_name
            if self.PYPI_API_TYPE == 'simple_html':
                package_canonical_name = canonicalize_name(package_name)
            response = requests.get(self.PYPI_API_URL.format(package=package_canonical_name), timeout=15)
        except HTTPError as e:  # pragma: nocover
            return False, e.message

        if not response.ok:  # pragma: nocover
            return False, 'API error: {}'.format(response.reason)

        if self.PYPI_API_TYPE == 'pypi_json':
            return self._parse_pypi_json_package_info(package_name, current_version, response)
        elif self.PYPI_API_TYPE == 'simple_html':
            return self._parse_simple_html_package_info(package_name, current_version, response)
        else:  # pragma: nocover
            raise NotImplementedError('This type of PYPI_API_TYPE type is not supported')

    def _expand_package(self, package_line):
        if '==' in package_line:
            name, vers = package_line.split('==')

            if '[' in name and name.strip().endswith(']'):
                name = name.split('[')[0]

            return name, vers

        return None, None

    def _parse_pypi_json_package_info(self, package_name, current_version, response):
        """
        :type package_name: str
        :type current_version: version.Version
        :type response: requests.models.Response
        """

        data = response.json()
        all_versions = [version.parse(vers) for vers in data['releases'].keys()]
        latest_version = max([vers for vers in all_versions if not vers.is_prerelease and not vers.is_postrelease])

        # even if user did not choose prerelease, if the package from requirements is pre/post release, use it
        if self._prerelease or current_version.is_postrelease or current_version.is_prerelease:
            prerelease_versions = [vers for vers in all_versions if vers.is_prerelease or vers.is_postrelease]
            if prerelease_versions:
                latest_version = max(prerelease_versions)
        try:
            try:
                latest_version_info = data['releases'][str(latest_version)][0]
            except KeyError:  # pragma: nocover
                # non-RFC versions, get the latest from pypi response
                latest_version = version.parse(data['info']['version'])
                latest_version_info = data['releases'][str(latest_version)][0]
        except Exception:  # pragma: nocover
            return False, 'error while parsing version'

        upload_time = latest_version_info['upload_time'].replace('T', ' ')

        return {
            'name': package_name,
            'current_version': current_version,
            'latest_version': latest_version,
            'upgrade_available': current_version < latest_version,
            'upload_time': upload_time
        }, 'success'

    def _parse_simple_html_package_info(self, package_name, current_version, response):
        """
        :type package_name: str
        :type current_version: version.Version
        :type response: requests.models.Response
        """
        pattern = r'<a.*>.*{name}-([A-z0-9\.-]*)(?:-py|\.tar).*<\/a>'.format(name=re.escape(package_name))
        versions_match = re.findall(pattern, response.content.decode('utf-8'), flags=re.IGNORECASE)

        all_versions = [version.parse(vers) for vers in versions_match]
        latest_version = max([vers for vers in all_versions if not vers.is_prerelease and not vers.is_postrelease])

        # even if user did not choose prerelease, if the package from requirements is pre/post release, use it
        if self._prerelease or current_version.is_postrelease or current_version.is_prerelease:
            prerelease_versions = [vers for vers in all_versions if vers.is_prerelease or vers.is_postrelease]
            if prerelease_versions:
                latest_version = max(prerelease_versions)

        return {
           'name': package_name,
           'current_version': current_version,
           'latest_version': latest_version,
           'upgrade_available': current_version < latest_version,
           'upload_time': '-'
        }, 'success'



"""
pip-upgrade

Usage:
  pip-upgrade [<requirements_file>] ... [--prerelease] [-p=<package>...] [--dry-run] [--skip-virtualenv-check] [--skip-package-installation] [--use-default-index]

Arguments:
    requirements_file             The requirement FILE, or WILDCARD PATH to multiple files.
    --prerelease                  Include prerelease versions for upgrade, when querying pypi repositories.
    -p <package>                  Pre-choose which packages tp upgrade. Skips any prompt.
    --dry-run                     Simulates the upgrade, but does not execute the actual upgrade.
    --skip-package-installation   Only upgrade the version in requirements files, don't install the new package.
    --skip-virtualenv-check       Disable virtualenv check. Allows installing the new packages outside the virtualenv.
    --use-default-index           Skip searching for custom index-url in pip configuration file(s).

Examples:
  pip-upgrade             # auto discovers requirements file
  pip-upgrade requirements.txt
  pip-upgrade requirements/dev.txt requirements/production.txt
  pip-upgrade requirements.txt -p django -p celery
  pip-upgrade requirements.txt -p all
  pip-upgrade requirements.txt --dry-run  # run everything as a simulation (don't do the actual upgrade)

Help:
  Interactively upgrade packages from requirements file, and also update the pinned version from requirements file(s).
  If no requirements are given, the command attempts to detect the requirements file(s) in the current directory.

  https://github.com/simion/pip-upgrader
"""  # noqa: E501
from __future__ import print_function, unicode_literals
from colorclass import Windows, Color
from docopt import docopt

from pip_upgrader import __version__ as VERSION
from pip_upgrader.packages_detector import PackagesDetector
from pip_upgrader.packages_interactive_selector import PackageInteractiveSelector
from pip_upgrader.packages_status_detector import PackagesStatusDetector
from pip_upgrader.packages_upgrader import PackagesUpgrader
from pip_upgrader.requirements_detector import RequirementsDetector
from pip_upgrader.virtualenv_checker import check_for_virtualenv


def get_options():
    return docopt(__doc__, version=VERSION)


def main():
    """ Main CLI entrypoint. """
    options = get_options()
    Windows.enable(auto_colors=True, reset_atexit=True)

    try:
        # maybe check if virtualenv is not activated
        check_for_virtualenv(options)

        # 1. detect requirements files
        filenames = RequirementsDetector(options.get('<requirements_file>')).get_filenames()
        if filenames:
            print(Color('{{autoyellow}}Found valid requirements file(s):{{/autoyellow}} '
                        '{{autocyan}}\n{}{{/autocyan}}'.format('\n'.join(filenames))))
        else:  # pragma: nocover
            print(Color('{autoyellow}No requirements files found in current directory. CD into your project '
                        'or manually specify requirements files as arguments.{/autoyellow}'))
            return
        # 2. detect all packages inside requirements
        packages = PackagesDetector(filenames).get_packages()

        # 3. query pypi API, see which package has a newer version vs the one in requirements (or current env)
        packages_status_map = PackagesStatusDetector(
            packages, options.get('--use-default-index')).detect_available_upgrades(options)

        # 4. [optionally], show interactive screen when user can choose which packages to upgrade
        selected_packages = PackageInteractiveSelector(packages_status_map, options).get_packages()

        # 5. having the list of packages, do the actual upgrade and replace the version inside all filenames
        upgraded_packages = PackagesUpgrader(selected_packages, filenames, options).do_upgrade()

        print(Color('{{autogreen}}Successfully upgraded (and updated requirements) for the following packages: '
                    '{}{{/autogreen}}'.format(','.join([package['name'] for package in upgraded_packages]))))
        if options['--dry-run']:
            print(Color('{automagenta}Actually, no, because this was a simulation using --dry-run{/automagenta}'))

    except KeyboardInterrupt:  # pragma: nocover
        print(Color('\n{autored}Upgrade interrupted.{/autored}'))



from __future__ import unicode_literals

from subprocess import PIPE, Popen as popen
from unittest import TestCase

import responses
from packaging.utils import canonicalize_name

from pip_upgrader import __version__ as VERSION

from pip_upgrader import cli
from pip_upgrader.packages_status_detector import PackagesStatusDetector

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

try:
    from io import StringIO
except ImportError:  # pragma: nocover
    from cStringIO import StringIO


class TestHelp(TestCase):
    def test_returns_usage_information(self):
        output = popen(['pip-upgrade', '-h'], stdout=PIPE).communicate()[0]
        self.assertTrue('Usage:' in output.decode('utf-8'))

        output = popen(['pip-upgrade', '--help'], stdout=PIPE).communicate()[0]
        self.assertTrue('Usage:' in output.decode('utf-8'))


class TestVersion(TestCase):
    def test_returns_version_information(self):
        output = popen(['pip-upgrade', '--version'], stdout=PIPE).communicate()[0]
        self.assertEqual(output.strip().decode('utf-8'), VERSION)


@patch('pip_upgrader.packages_interactive_selector.user_input', return_value='all')
@patch('pip_upgrader.virtualenv_checker.is_virtualenv', return_value=True)
class TestCommand(TestCase):

    def _add_responses_mocks(self):
        for package in ['Django', 'celery', 'django-rest-auth', 'ipython']:
            with open('tests/fixtures/{}.json'.format(package)) as fh:
                body = fh.read()

            responses.add(responses.GET,
                          "https://pypi.python.org/pypi/{}/json".format(package),
                          body=body,
                          content_type="application/json")

            with open('tests/fixtures/{}.html'.format(canonicalize_name(package))) as fh:
                body_html = fh.read()
            responses.add(responses.GET,
                          "https://pypi.python.org/simple/{}".format(canonicalize_name(package)),
                          body=body_html)

    def setUp(self):
        self._add_responses_mocks()

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '-p': []})
    def test_command_basic_usage(self, options_mock, is_virtualenv_mock, user_input_mock):

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            cli.main()
            output = stdout_mock.getvalue()

        self.assertTrue(user_input_mock.called)

        self.assertIn('Available upgrades', output)
        self.assertIn('ipython ... up to date', output)
        self.assertIn('django-rest-auth ... upgrade available: 0.9.0 ==>', output)
        self.assertNotIn('ipdb', output)
        self.assertIn('Successfully upgraded', output)
        self.assertIn('this was a simulation using --dry-run', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '-p': []})
    def test_command_simple_html_index_url(self, options_mock, is_virtualenv_mock, user_input_mock):

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock, \
                patch('pip_upgrader.packages_status_detector.PackagesStatusDetector.pip_config_locations',
                      new=PackagesStatusDetector.pip_config_locations + ['pip.test.conf']):
            cli.main()
            output = stdout_mock.getvalue()

        self.assertTrue(user_input_mock.called)
        # checks if new index-url was discovered from config file
        self.assertIn('Setting API url', output)
        self.assertIn('https://pypi.python.org/simple/{package}', output)

        self.assertIn('Available upgrades', output)
        self.assertIn('ipython ... up to date', output)
        self.assertIn('django-rest-auth ... upgrade available: 0.9.0 ==>', output)
        self.assertNotIn('ipdb', output)
        self.assertIn('Successfully upgraded', output)
        self.assertIn('this was a simulation using --dry-run', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '-p': [], '--use-default-index': True})
    def test_command__use_default_index(self, options_mock, is_virtualenv_mock, user_input_mock):

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock, \
                patch('pip_upgrader.packages_status_detector.PackagesStatusDetector.pip_config_locations',
                      new=PackagesStatusDetector.pip_config_locations + ['pip.test.conf']):
            cli.main()
            output = stdout_mock.getvalue()

        # checks if new index-url was discovered from config file
        self.assertNotIn('Setting API url', output)
        self.assertIn('Successfully upgraded', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '-p': []})
    def test_command_interactive_bad_choices(self, options_mock, is_virtualenv_mock, user_input_mock):

        user_input_mock.return_value = ''
        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            cli.main()
            output = stdout_mock.getvalue()

        self.assertTrue(user_input_mock.called)
        self.assertIn('No choice selected', output)
        self.assertNotIn('Setting API url', output)

        user_input_mock.return_value = '5 6 7'
        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            cli.main()
            output = stdout_mock.getvalue()

        self.assertTrue(user_input_mock.called)
        self.assertIn('No valid choice selected.', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '-p': ['all']})
    def test_command_not_interactive_all_packages(self, options_mock, is_virtualenv_mock, user_input_mock):

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            cli.main()
            output = stdout_mock.getvalue()

        # no user_input should be called
        self.assertFalse(user_input_mock.called)

        self.assertNotIn('Setting API url', output)
        self.assertNotIn('Available upgrades', output)
        self.assertIn('Django ... upgrade available: 1.10 ==>', output)
        self.assertIn('django-rest-auth ... upgrade available: 0.9.0 ==>', output)
        self.assertIn('ipython ... up to date', output)
        self.assertNotIn('ipdb', output)
        self.assertIn('celery ... upgrade available: 3.1.1 ==>', output)

        self.assertIn('Successfully upgraded', output)
        self.assertIn('this was a simulation using --dry-run', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '-p': ['django', 'bad_package']})
    def test_command_not_interactive_specific_package(self, options_mock, is_virtualenv_mock, user_input_mock):

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            cli.main()
            output = stdout_mock.getvalue()

        # no user_input should be called
        self.assertFalse(user_input_mock.called)

        self.assertNotIn('Setting API url', output)
        self.assertIn('Django ... upgrade available: 1.10 ==>', output)
        self.assertNotIn('django-rest-auth', output)
        self.assertNotIn('ipython ... up to date', output)
        self.assertNotIn('ipdb', output)
        self.assertNotIn('celery ... upgrade available: 3.1.1 ==>', output)

        self.assertIn('Successfully upgraded', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '-p': ['ipython']})
    def test_command_not_interactive_all_packages_up_to_date(self, options_mock, is_virtualenv_mock, user_input_mock):

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            cli.main()
            output = stdout_mock.getvalue()

        # no user_input should be called
        self.assertFalse(user_input_mock.called)
        self.assertNotIn('Setting API url', output)
        self.assertIn('All packages are up-to-date.', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '-p': ['all'],
                                                         '<requirements_file>': ['requirements/production.txt']})
    def test_command_not_interactive_explicit_requirements(self, options_mock, is_virtualenv_mock, user_input_mock):

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            cli.main()
            output = stdout_mock.getvalue()

        # no user_input should be called
        self.assertFalse(user_input_mock.called)

        self.assertNotIn('Setting API url', output)
        self.assertNotIn('Django ... upgrade available: 1.10 ==>', output)
        self.assertNotIn('django-rest-auth', output)
        self.assertNotIn('ipython ... up to date', output)
        self.assertNotIn('ipdb', output)
        self.assertIn('celery ... upgrade available: 3.1.1 ==>', output)

        self.assertIn('Successfully upgraded', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '-p': ['django'], '--prerelease': True})
    def test_command_not_specific_package_prerelease(self, options_mock, is_virtualenv_mock, user_input_mock):

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            cli.main()
            output = stdout_mock.getvalue()

        # no user_input should be called
        self.assertFalse(user_input_mock.called)

        self.assertNotIn('Setting API url', output)
        self.assertIn('Django ... upgrade available: 1.10 ==> 1.11rc1', output)
        self.assertNotIn('django-rest-auth', output)
        self.assertNotIn('ipython ... up to date', output)
        self.assertNotIn('ipdb', output)
        self.assertNotIn('celery ... upgrade available: 3.1.1 ==>', output)

        self.assertIn('Successfully upgraded', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '-p': ['django'], '--prerelease': True})
    def test_command_not_specific_package_prerelease_html_api(self, options_mock, is_virtualenv_mock, user_input_mock):

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock, \
                patch('pip_upgrader.packages_status_detector.PackagesStatusDetector.pip_config_locations',
                      new=PackagesStatusDetector.pip_config_locations + ['pip.test.conf']):
            cli.main()
            output = stdout_mock.getvalue()

        # no user_input should be called
        self.assertFalse(user_input_mock.called)

        self.assertIn('Setting API url', output)
        self.assertIn('Django ... upgrade available: 1.10 ==> 1.11rc1', output)
        self.assertNotIn('django-rest-auth', output)
        self.assertNotIn('ipython ... up to date', output)
        self.assertNotIn('ipdb', output)
        self.assertNotIn('celery ... upgrade available: 3.1.1 ==>', output)

        self.assertIn('Successfully upgraded', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '--skip-virtualenv-check': False,
                                                         '-p': ['django']})
    def test_command_not_interactive_not_virtualenv(self, options_mock, is_virtualenv_mock, user_input_mock):
        is_virtualenv_mock.return_value = False

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            cli.main()
            output = stdout_mock.getvalue()

        self.assertNotIn('Setting API url', output)
        self.assertIn("It seems you haven't activated a virtualenv", output)
        self.assertNotIn('Successfully upgraded', output)

    @responses.activate
    @patch('pip_upgrader.cli.get_options', return_value={'--dry-run': True, '--skip-virtualenv-check': True,
                                                         '-p': ['django']})
    def test_command_not_interactive_not_virtualenv_skip(self, options_mock, is_virtualenv_mock, user_input_mock):
        is_virtualenv_mock.return_value = False

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            cli.main()
            output = stdout_mock.getvalue()

        self.assertFalse(user_input_mock.called)
        self.assertNotIn('Setting API url', output)
        self.assertIn('Django ... upgrade available: 1.10 ==>', output)
        self.assertNotIn('django-rest-auth', output)
        self.assertNotIn('ipython ... up to date', output)
        self.assertNotIn('ipdb', output)
        self.assertNotIn('celery ... upgrade available: 3.1.1 ==>', output)
        self.assertIn('Successfully upgraded', output)



from __future__ import unicode_literals, print_function
import os
import sys

from colorclass import Color


def is_virtualenv():  # pragma: nocover (this is mocked, cannot be tested)
    if getattr(sys, 'base_prefix', sys.prefix) != sys.prefix:
        return True

    if hasattr(sys, 'real_prefix'):
        return True

    if os.environ.get('VIRTUAL_ENV'):
        return True

    return False


def check_for_virtualenv(options):
    if options.get('--skip-virtualenv-check', False) or options.get('--skip-package-installation', False):
        return  # no check needed

    if not is_virtualenv():
        print(Color("{autoyellow}It seems you haven't activated a virtualenv. \n"
                    "Installing packages directly in the system is not recommended. \n"
                    "{automagenta}Activate your project's virtualenv{/automagenta}, "
                    "or {automagenta}re-run this command {/automagenta}with one of the following options:\n"
                    "--skip-virtualenv-check (install the packages anyway)\n"
                    "--skip-package-installation (don't install any package. just update the requirements file(s))"
                    "{/autoyellow}"))
        raise KeyboardInterrupt()



from __future__ import print_function, unicode_literals

from collections import OrderedDict

from colorclass import Color
from terminaltables import AsciiTable


def user_input(prompt=None):  # pragma: nocover
    try:
        input_func = raw_input
    except NameError:
        input_func = input
    return input_func(prompt)


class PackageInteractiveSelector(object):
    packages_for_upgrade = OrderedDict()
    selected_packages = []

    def __init__(self, packages_map, options):
        self.selected_packages = []
        self.packages_for_upgrade = {}

        # map with index number, for later choosing
        i = 1
        for package in packages_map.values():
            if package['upgrade_available']:
                self.packages_for_upgrade[i] = package.copy()
                i += 1

        # maybe all packages are up-to-date
        if not self.packages_for_upgrade:
            print(Color('{autogreen}All packages are up-to-date.{/autogreen}'))
            raise KeyboardInterrupt()

        # choose which packages to upgrade (interactive or not)
        if '-p' in options and options['-p']:
            if options['-p'] == ['all']:
                self._select_packages(self.packages_for_upgrade.keys())
            else:
                for index, package in self.packages_for_upgrade.items():
                    for chosen_package in options['-p']:
                        if chosen_package.lower().strip() == package['name'].lower().strip():
                            self._select_packages([index])
        else:
            self.ask_for_packages()

    def get_packages(self):
        return self.selected_packages

    def ask_for_packages(self):
        data = [[
            Color('{autoblue}No.{/autoblue}'),
            Color('{autoblue}Package{/autoblue}'),
            Color('{autoblue}Current version{/autoblue}'),
            Color('{autoblue}Latest version{/autoblue}'),
            Color('{autoblue}Release date{/autoblue}'),
        ]]

        for i, package in self.packages_for_upgrade.items():
            data.append([Color('{{autobgblack}}{{autogreen}} {} {{/autogreen}}{{/bgblack}}'.format(i)),
                         Color('{{autogreen}} {} {{/autogreen}}'.format(package['name'])),
                         package['current_version'],
                         package['latest_version'],
                         package['upload_time']])

        print('')
        print(Color('{autogreen}Available upgrades:{/autogreen}'))
        table = AsciiTable(data)
        print(table.table)
        print('')

        print('Please choose which packages should be upgraded. Choices: "all", "q" (quit), "x" (exit) or "1 2 3"')
        choice = user_input(Color('{autogreen}Choice:{/autogreen} ')).strip()

        if not choice and not choice.strip():
            print(Color('{autored}No choice selected.{/autored}'))
            raise KeyboardInterrupt()

        choice = choice.strip()

        if choice == 'q':  # pragma: nocover
            print(Color('{autored}Quit.{/autored}'))
            raise KeyboardInterrupt()

        if choice == 'x':  # pragma: nocover
            print(Color('{autored}Exit.{/autored}'))
            raise KeyboardInterrupt()

        if choice == "all":
            self._select_packages(self.packages_for_upgrade.keys())
        else:
            try:
                selected = list(self._select_packages([int(index.strip()) for index in choice.split(' ')]))
                if not any(selected):
                    print(Color('{autored}No valid choice selected.{/autored}'))
                    raise KeyboardInterrupt()
            except ValueError:  # pragma: nocover
                print(Color('{autored}Invalid choice{/autored}'))
                raise KeyboardInterrupt()

    def _select_packages(self, indexes):
        selected = []

        for index in indexes:
            if index in self.packages_for_upgrade:
                self.selected_packages.append(self.packages_for_upgrade[index].copy())
                selected.append(True)

        return selected



"""Packaging settings."""


from codecs import open
from os.path import abspath, dirname, join
from subprocess import call

from setuptools import Command, find_packages, setup


this_dir = abspath(dirname(__file__))
with open(join(this_dir, 'README.rst'), encoding='utf-8') as file:
    long_description = file.read()


class RunTests(Command):
    """ Run all tests. """
    description = 'run tests'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run all tests!"""
        err = call(['py.test'])
        raise SystemExit(err)


setup(
    name='pip_upgrader',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    description='An interactive pip requirements upgrader. It also updates the version in your requirements.txt file.',
    long_description=long_description,
    url='https://github.com/simion/pip-upgrader',
    author='Simion Baws',
    author_email='simion.agv@gmail.com',
    license='Apache License 2.0',
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'License :: Public Domain',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='cli,pip,pypi,requirements,upgrade',
    packages=find_packages(exclude=['docs', 'tests*']),
    install_requires=['docopt', 'packaging', 'requests', 'terminaltables', 'colorclass'],
    extras_require={
        'test': ['coverage', 'pytest', 'pytest-cov', 'pytest-pep8', 'mock', 'responses'],
    },
    entry_points={
        'console_scripts': [
            'pip-upgrade=pip_upgrader.cli:main',
        ],
    },
    cmdclass={'test': RunTests},
)
